<?php

/**
 * Helper class.
 */
class TmgmtCdtHelper {

  const SIMULATION = 'SIMULATION';
  const PRODUCTION = 'PRODUCTION';

  const NAME = 'Translation Centre (CdT)';
  const DEFAULT_REQUEST_IDENTIFIER = '0000/unknown';
  const TOKEN_TYPE_DRUPAL_TOKEN_MEDIA = 'drupal_token_media';
  const TOKEN_TYPE_DRUPAL_TOKEN_SIMPLE = 'drupal_token_simple';
  const TOKEN_TYPE_HTML_A = 'html_a';
  const TOKEN_TYPE_HTML_IMG = 'html_img';
  const TOKEN_TYPE_HTML_SCRIPT = 'html_script';

  const API_STATUS_COMPLETED = 'COMPLETED';
  const API_STATUS_QUOTED_PENDING_APPROVAL = 'QUOTED - PENDING APPROVAL';
  const API_STATUS_CANCELLED = 'CANCELLED';

  const TRANSLATABLE_TYPES = 'a:6:{i:0;s:4:"node";i:1;s:13:"taxonomy_term";i:2;s:9:"menu_link";i:3;s:5:"block";i:4;s:5:"views";i:5;s:7:"default";}';

  /**
   * Check if is a json.
   */
  public static function isJson($data) {
        // @codingStandardsIgnoreStart
        // Need native json format.
        if (is_string($data) && is_array(\json_decode($data, true)) && (json_last_error() === JSON_ERROR_NONE)) {
            return true;
        } else {
            return false;
        }
        // @codingStandardsIgnoreEnd
  }

  /**
   * WatchdogDebug.
   */
  public static function watchdogDebug($var, $label = 'label') {

    $bt = debug_backtrace();
    $caller = array_shift($bt);

    watchdog(
          $label,
          '<pre>@file: @line</pre>@label: <pre>@var</pre>',
          array(
            '@label' => $label,
            '@var' => print_r($var, TRUE),
            '@file' => $caller['file'],
            '@line' => $caller['line'],
          ),
          WATCHDOG_DEBUG
      );
  }

  /**
   * File_get_contents curl version (to use same proxy params as api)
   */
  public static function curlGetFileContents($url) {
    try {

      $proxy_host = variable_get('tmgmt_cdt_curl_proxy_host');
      $proxy_port = variable_get('tmgmt_cdt_curl_proxy_port');
      $proxy_user = variable_get('tmgmt_cdt_curl_proxy_user');
      $proxy_pass = variable_get('tmgmt_cdt_curl_proxy_pass');
      $proxy_user_agent = variable_get('tmgmt_cdt_curl_proxy_user_agent');
      $curl_options = drupal_json_decode(variable_get('tmgmt_cdt_curl_options'));

      $curl = curl_init();
      $curlopt[CURLOPT_RETURNTRANSFER] = 1;
      $curlopt[CURLOPT_URL] = $url;

      if (!empty($proxy_host)) {
        $curlopt[CURLOPT_PROXY] = $proxy_host;
        $curlopt[CURLOPT_PROXYPORT] = $proxy_port;
      }

      if (!empty($proxy_user)) {
        $curlopt[CURLOPT_PROXYUSERPWD] = $proxy_user . ':' . $proxy_pass;
      }

      if (!empty($proxy_user_agent)) {
        $curlopt[CURLOPT_USERAGENT] = $proxy_user_agent;
      }

      foreach ($curl_options as $curl_option => $curl_value) {
        $curlopt[constant($curl_option)] = $curl_value;
      }

      curl_setopt_array($curl, $curlopt);

      $raw_data = curl_exec($curl);

      if (curl_errno($curl)) {
        throw new \Exception(curl_error($curl));
      }

      curl_close($curl);
      unset($curl);

      if ($raw_data) {
        return $raw_data;
      }
      else {
        return FALSE;
      }

    }
    catch (\Exception $e) {
      drupal_set_message($e->getMessage(), 'error', FALSE);
      return FALSE;
    }

  }

  /**
   * All fields used in this module.
   */
  public static function getFields() {
    return array(
      'tmgmt_job' => array(
        'cdt_correlation_id' => array(
          'label' => 'Correlation ID',
          'description' => 'ID returned after a successful request to CDT.',
          'db' => array(
            'type' => 'varchar',
            'length' => 32,
          ),
          'entity' => array(
            'type' => 'text',
          ),
        ),
        'cdt_request_identifier' => array(
          'label' => 'Request identifier',
          'description' => 'Seconde ID returned by CDT.',
          'db' => array(
            'type' => 'varchar',
            'length' => 32,
          ),
          'entity' => array(
            'type' => 'text',
          ),
        ),
        'cdt_request_params' => array(
          'label' => 'Request parameters',
          'description' => 'Parameters used to make the request to CDT.',
          'db' => array(
            'type' => 'text',
            'size' => 'medium',
          ),
          'entity' => array(
            'type' => 'text',
          ),
        ),
        'cdt_response' => array(
          'label' => 'Response',
          'description' => 'Response (translations) from CDT.',
          'db' => array(
            'type' => 'text',
            'size' => 'medium',
          ),
          'entity' => array(
            'type' => 'text',
          ),
        ),
      ),
      'tmgmt_job_item' => array(
        'cdt_source_language' => array(
          'label' => 'Source language',
          'description' => 'Source language.',
          'db' => array(
            'type' => 'varchar',
            'length' => 12,
          ),
          'entity' => array(
            'type' => 'text',
          ),
        ),
        'cdt_target_language' => array(
          'label' => 'Target language',
          'description' => 'Target language.',
          'db' => array(
            'type' => 'varchar',
            'length' => 12,
          ),
          'entity' => array(
            'type' => 'text',
          ),
        ),
        'cdt_character_count' => array(
          'label' => 'Character count',
          'description' => 'Character count.',
          'db' => array(
            'type' => 'int',
          ),
          'entity' => array(
            'type' => 'integer',
          ),
        ),
        'cdt_settings' => array(
          'label' => 'Settings',
          'description' => 'Settings.',
          'db' => array(
            'type' => 'text',
            'size' => 'medium',
            'serialize' => TRUE,
          ),
          'entity' => array(
            'type' => 'text',
          ),

        ),
      ),
    );
  }

  /**
   * Get default options.
   */
  public static function defaultOptions() {

    $options = [

      'tmgmt_cdt_translatable_fields' => serialize(array()),
      'tmgmt_cdt_api_mode' => self::SIMULATION,
      'tmgmt_cdt_ready' => FALSE,
      'tmgmt_cdt_debug' => FALSE,

      'tmgmt_cdt_business_reference_data' => new \stdClass(),
      'tmgmt_cdt_api_url' => '',
      'tmgmt_cdt_api_username' => '',
      'tmgmt_cdt_api_password' => '',

      'tmgmt_cdt_curl_proxy_host' => '',
      'tmgmt_cdt_curl_proxy_port' => '',
      'tmgmt_cdt_curl_proxy_user' => '',
      'tmgmt_cdt_curl_proxy_pass' => '',
      'tmgmt_cdt_curl_proxy_user_agent' => '',
      'tmgmt_cdt_curl_options' => '{"CURLOPT_TIMEOUT":15,"CURLOPT_RETURNTRANSFER":true,"CURLOPT_SSL_VERIFYHOST":0,"CURLOPT_SSL_VERIFYPEER":0}',

      'tmgmt_cdt_max_pages_per_request' => 10,
      'tmgmt_cdt_use_token_html_a' => FALSE,
      'tmgmt_cdt_use_token_html_img' => FALSE,
      'tmgmt_cdt_use_token_html_script' => TRUE,
      'tmgmt_cdt_use_token_drupal' => TRUE,
      'tmgmt_cdt_html_check' => TRUE,
      'tmgmt_cdt_xsd_link' => 'http://static.cdt.europa.eu/webtranslations/schemas/Drupal-Translation-V8-0.xsd',

      'tmgmt_cdt_disable_xsd_validation' => FALSE,
      'tmgmt_cdt_show_quotation' => FALSE,
      'tmgmt_cdt_xsd_source' => '',

      'tmgmt_cdt_api_access_token_expire' => 0,
      'tmgmt_cdt_api_access_token' => '',

    ];

    return $options;

  }

  /**
   * Show if the module status is ready or not.
   */
  public static function showStatus($showReady = FALSE) {
    if ($showReady) {
      if (!(bool) variable_get('tmgmt_cdt_ready')) {
        drupal_set_message(t('<strong>@name</strong> translator is <strong>not ready</strong>.', ['@name' => self::NAME]), 'warning');
      }
      else {
        drupal_set_message(t('<strong>@name</strong> translator is <strong>ready</strong>.', ['@name' => self::NAME]), 'status');
      }
    }

    if (variable_get('tmgmt_cdt_api_mode') === self::SIMULATION && (bool) variable_get('tmgmt_cdt_ready')) {
      drupal_set_message(t('<strong>@name</strong> use the mode <strong>@mode</strong>.',
            [
              '@name' => self::NAME,
              '@mode' => variable_get('tmgmt_cdt_api_mode'),
            ]), 'warning');
    }

  }

  /**
   * Extract year and number.
   */
  public static function splitRequestIdentifier($request_number) {
    $request_ids = explode('/', trim($request_number));
    $request_year = @trim($request_ids[0]);
    $request_number = @trim($request_ids[1]);
    return [
      'requestyear' => $request_year,
      'requestnumber' => $request_number,
    ];
  }

  /**
   * Return job by requestidentifier.
   */
  public static function jobByRequestIdentifier($request_identifier) {
    if ($request_identifier === self::DEFAULT_REQUEST_IDENTIFIER) {
      return FALSE;
    }

    $query = new EntityFieldQuery();
    $query
      ->entityCondition('entity_type', 'tmgmt_job')
      ->propertyCondition('translator', 'cdt')
      ->propertyCondition('cdt_request_identifier', $request_identifier);
    $result = $query->execute();
    if (empty($result['tmgmt_job'])) {
      return FALSE;
    }

    $ids = array_keys($result['tmgmt_job']);
    $job = entity_load_single('tmgmt_job', reset($ids));

    return $job;
  }

}
